@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-
 # This file is part of beets.
 # Copyright 2020, Callum Brown.
 #
@@ -16,36 +14,41 @@
 
 """An AURA server using Flask."""
 
-from __future__ import division, absolute_import, print_function
-
-from mimetypes import guess_type
+import os
 import re
-from os.path import isfile, getsize
-
-from beets.plugins import BeetsPlugin
-from beets.ui import Subcommand, _open_library
-from beets import config
-from beets.util import py3_path
-from beets.library import Item, Album
-from beets.dbcore.query import (
-    MatchQuery,
-    NotQuery,
-    RegexpQuery,
-    AndQuery,
-    FixedFieldSort,
-    SlowFieldSort,
-    MultipleSort,
-)
+import sys
+from dataclasses import dataclass
+from mimetypes import guess_type
+from typing import ClassVar, Mapping, Type
 
 from flask import (
     Blueprint,
     Flask,
     current_app,
-    send_file,
     make_response,
     request,
+    send_file,
 )
 
+if sys.version_info >= (3, 11):
+    from typing import Self
+else:
+    from typing_extensions import Self
+
+from beets import config
+from beets.dbcore.query import (
+    AndQuery,
+    FixedFieldSort,
+    MatchQuery,
+    MultipleSort,
+    NotQuery,
+    RegexpQuery,
+    SlowFieldSort,
+    SQLiteType,
+)
+from beets.library import Album, Item, LibModel, Library
+from beets.plugins import BeetsPlugin
+from beets.ui import Subcommand, _open_library
 
 # Constants
 
@@ -120,9 +123,20 @@ ARTIST_ATTR_MAP = {
 }
 
 
+@dataclass
 class AURADocument:
     """Base class for building AURA documents."""
 
+    model_cls: ClassVar[Type[LibModel]]
+
+    lib: Library
+    args: Mapping[str, str]
+
+    @classmethod
+    def from_app(cls) -> Self:
+        """Initialise the document using the global app and request."""
+        return cls(current_app.config["lib"], request.args)
+
     @staticmethod
     def error(status, title, detail):
         """Make a response for an error following the JSON:API spec.
@@ -138,13 +152,29 @@ class AURADocument:
         }
         return make_response(document, status)
 
+    @classmethod
+    def get_attribute_converter(cls, beets_attr: str) -> Type[SQLiteType]:
+        """Work out what data type an attribute should be for beets.
+
+        Args:
+            beets_attr: The name of the beets attribute, e.g. "title".
+        """
+        try:
+            # Look for field in list of Album fields
+            # and get python type of database type.
+            # See beets.library.Album and beets.dbcore.types
+            return cls.model_cls._fields[beets_attr].model_type
+        except KeyError:
+            # Fall back to string (NOTE: probably not good)
+            return str
+
     def translate_filters(self):
         """Translate filters from request arguments to a beets Query."""
         # The format of each filter key in the request parameter is:
         # filter[<attribute>]. This regex extracts <attribute>.
         pattern = re.compile(r"filter\[(?P<attribute>[a-zA-Z0-9_-]+)\]")
         queries = []
-        for key, value in request.args.items():
+        for key, value in self.args.items():
             match = pattern.match(key)
             if match:
                 # Extract attribute name from key
@@ -193,10 +223,10 @@ class AURADocument:
                 albums) or a list of strings (artists).
         """
         # Pages start from zero
-        page = request.args.get("page", 0, int)
+        page = self.args.get("page", 0, int)
         # Use page limit defined in config by default.
         default_limit = config["aura"]["page_limit"].get(int)
-        limit = request.args.get("limit", default_limit, int)
+        limit = self.args.get("limit", default_limit, int)
         # start = offset of first item to return
         start = page * limit
         # end = offset of last item + 1
@@ -206,19 +236,22 @@ class AURADocument:
             next_url = None
         else:
             # Not the last page so work out links.next url
-            if not request.args:
+            if not self.args:
                 # No existing arguments, so current page is 0
                 next_url = request.url + "?page=1"
-            elif not request.args.get("page", None):
+            elif not self.args.get("page", None):
                 # No existing page argument, so add one to the end
                 next_url = request.url + "&page=1"
             else:
                 # Increment page token by 1
                 next_url = request.url.replace(
-                    "page={}".format(page), "page={}".format(page + 1)
+                    f"page={page}", "page={}".format(page + 1)
                 )
         # Get only the items in the page range
-        data = [self.resource_object(collection[i]) for i in range(start, end)]
+        data = [
+            self.get_resource_object(self.lib, collection[i])
+            for i in range(start, end)
+        ]
         return data, next_url
 
     def get_included(self, data, include_str):
@@ -252,26 +285,34 @@ class AURADocument:
             res_type = identifier["type"]
             if res_type == "track":
                 track_id = int(identifier["id"])
-                track = current_app.config["lib"].get_item(track_id)
-                included.append(TrackDocument.resource_object(track))
+                track = self.lib.get_item(track_id)
+                included.append(
+                    TrackDocument.get_resource_object(self.lib, track)
+                )
             elif res_type == "album":
                 album_id = int(identifier["id"])
-                album = current_app.config["lib"].get_album(album_id)
-                included.append(AlbumDocument.resource_object(album))
+                album = self.lib.get_album(album_id)
+                included.append(
+                    AlbumDocument.get_resource_object(self.lib, album)
+                )
             elif res_type == "artist":
                 artist_id = identifier["id"]
-                included.append(ArtistDocument.resource_object(artist_id))
+                included.append(
+                    ArtistDocument.get_resource_object(self.lib, artist_id)
+                )
             elif res_type == "image":
                 image_id = identifier["id"]
-                included.append(ImageDocument.resource_object(image_id))
+                included.append(
+                    ImageDocument.get_resource_object(self.lib, image_id)
+                )
             else:
-                raise ValueError("Invalid resource type: {}".format(res_type))
+                raise ValueError(f"Invalid resource type: {res_type}")
         return included
 
     def all_resources(self):
         """Build document for /tracks, /albums or /artists."""
         query = self.translate_filters()
-        sort_arg = request.args.get("sort", None)
+        sort_arg = self.args.get("sort", None)
         if sort_arg:
             sort = self.translate_sorts(sort_arg)
             # For each sort field add a query which ensures all results
@@ -294,7 +335,7 @@ class AURADocument:
         if next_url:
             document["links"] = {"next": next_url}
         # Include related resources for each element in "data"
-        include_str = request.args.get("include", None)
+        include_str = self.args.get("include", None)
         if include_str:
             document["included"] = self.get_included(data, include_str)
         return document
@@ -307,7 +348,7 @@ class AURADocument:
                 resource object.
         """
         document = {"data": resource_object}
-        include_str = request.args.get("include", None)
+        include_str = self.args.get("include", None)
         if include_str:
             # [document["data"]] is because arg needs to be list
             document["included"] = self.get_included(
@@ -319,6 +360,8 @@ class AURADocument:
 class TrackDocument(AURADocument):
     """Class for building documents for /tracks endpoints."""
 
+    model_cls = Item
+
     attribute_map = TRACK_ATTR_MAP
 
     def get_collection(self, query=None, sort=None):
@@ -328,9 +371,10 @@ class TrackDocument(AURADocument):
             query: A beets Query object or a beets query string.
             sort: A beets Sort object.
         """
-        return current_app.config["lib"].items(query, sort)
+        return self.lib.items(query, sort)
 
-    def get_attribute_converter(self, beets_attr):
+    @classmethod
+    def get_attribute_converter(cls, beets_attr: str) -> Type[SQLiteType]:
         """Work out what data type an attribute should be for beets.
 
         Args:
@@ -338,20 +382,12 @@ class TrackDocument(AURADocument):
         """
         # filesize is a special field (read from disk not db?)
         if beets_attr == "filesize":
-            converter = int
-        else:
-            try:
-                # Look for field in list of Item fields
-                # and get python type of database type.
-                # See beets.library.Item and beets.dbcore.types
-                converter = Item._fields[beets_attr].model_type
-            except KeyError:
-                # Fall back to string (NOTE: probably not good)
-                converter = str
-        return converter
+            return int
+
+        return super().get_attribute_converter(beets_attr)
 
     @staticmethod
-    def resource_object(track):
+    def get_resource_object(lib: Library, track):
         """Construct a JSON:API resource object from a beets Item.
 
         Args:
@@ -389,7 +425,7 @@ class TrackDocument(AURADocument):
         Args:
             track_id: The beets id of the track (integer).
         """
-        track = current_app.config["lib"].get_item(track_id)
+        track = self.lib.get_item(track_id)
         if not track:
             return self.error(
                 "404 Not Found",
@@ -398,12 +434,16 @@ class TrackDocument(AURADocument):
                     track_id
                 ),
             )
-        return self.single_resource_document(self.resource_object(track))
+        return self.single_resource_document(
+            self.get_resource_object(self.lib, track)
+        )
 
 
 class AlbumDocument(AURADocument):
     """Class for building documents for /albums endpoints."""
 
+    model_cls = Album
+
     attribute_map = ALBUM_ATTR_MAP
 
     def get_collection(self, query=None, sort=None):
@@ -413,26 +453,10 @@ class AlbumDocument(AURADocument):
             query: A beets Query object or a beets query string.
             sort: A beets Sort object.
         """
-        return current_app.config["lib"].albums(query, sort)
-
-    def get_attribute_converter(self, beets_attr):
-        """Work out what data type an attribute should be for beets.
-
-        Args:
-            beets_attr: The name of the beets attribute, e.g. "title".
-        """
-        try:
-            # Look for field in list of Album fields
-            # and get python type of database type.
-            # See beets.library.Album and beets.dbcore.types
-            converter = Album._fields[beets_attr].model_type
-        except KeyError:
-            # Fall back to string (NOTE: probably not good)
-            converter = str
-        return converter
+        return self.lib.albums(query, sort)
 
     @staticmethod
-    def resource_object(album):
+    def get_resource_object(lib: Library, album):
         """Construct a JSON:API resource object from a beets Album.
 
         Args:
@@ -451,7 +475,7 @@ class AlbumDocument(AURADocument):
         # track number. Sorting is not required but it's nice.
         query = MatchQuery("album_id", album.id)
         sort = FixedFieldSort("track", ascending=True)
-        tracks = current_app.config["lib"].items(query, sort)
+        tracks = lib.items(query, sort)
         # JSON:API one-to-many relationship to tracks on the album
         relationships = {
             "tracks": {
@@ -460,9 +484,9 @@ class AlbumDocument(AURADocument):
         }
         # Add images relationship if album has associated images
         if album.artpath:
-            path = py3_path(album.artpath)
+            path = os.fsdecode(album.artpath)
             filename = path.split("/")[-1]
-            image_id = "album-{}-{}".format(album.id, filename)
+            image_id = f"album-{album.id}-{filename}"
             relationships["images"] = {
                 "data": [{"type": "image", "id": image_id}]
             }
@@ -487,7 +511,7 @@ class AlbumDocument(AURADocument):
         Args:
             album_id: The beets id of the album (integer).
         """
-        album = current_app.config["lib"].get_album(album_id)
+        album = self.lib.get_album(album_id)
         if not album:
             return self.error(
                 "404 Not Found",
@@ -496,12 +520,16 @@ class AlbumDocument(AURADocument):
                     album_id
                 ),
             )
-        return self.single_resource_document(self.resource_object(album))
+        return self.single_resource_document(
+            self.get_resource_object(self.lib, album)
+        )
 
 
 class ArtistDocument(AURADocument):
     """Class for building documents for /artists endpoints."""
 
+    model_cls = Item
+
     attribute_map = ARTIST_ATTR_MAP
 
     def get_collection(self, query=None, sort=None):
@@ -512,7 +540,7 @@ class ArtistDocument(AURADocument):
             sort: A beets Sort object.
         """
         # Gets only tracks with matching artist information
-        tracks = current_app.config["lib"].items(query, sort)
+        tracks = self.lib.items(query, sort)
         collection = []
         for track in tracks:
             # Do not add duplicates
@@ -520,24 +548,8 @@ class ArtistDocument(AURADocument):
                 collection.append(track.artist)
         return collection
 
-    def get_attribute_converter(self, beets_attr):
-        """Work out what data type an attribute should be for beets.
-
-        Args:
-            beets_attr: The name of the beets attribute, e.g. "artist".
-        """
-        try:
-            # Look for field in list of Item fields
-            # and get python type of database type.
-            # See beets.library.Item and beets.dbcore.types
-            converter = Item._fields[beets_attr].model_type
-        except KeyError:
-            # Fall back to string (NOTE: probably not good)
-            converter = str
-        return converter
-
     @staticmethod
-    def resource_object(artist_id):
+    def get_resource_object(lib: Library, artist_id):
         """Construct a JSON:API resource object for the given artist.
 
         Args:
@@ -545,7 +557,7 @@ class ArtistDocument(AURADocument):
         """
         # Get tracks where artist field exactly matches artist_id
         query = MatchQuery("artist", artist_id)
-        tracks = current_app.config["lib"].items(query)
+        tracks = lib.items(query)
         if not tracks:
             return None
 
@@ -567,7 +579,7 @@ class ArtistDocument(AURADocument):
             }
         }
         album_query = MatchQuery("albumartist", artist_id)
-        albums = current_app.config["lib"].albums(query=album_query)
+        albums = lib.albums(query=album_query)
         if len(albums) != 0:
             relationships["albums"] = {
                 "data": [{"type": "album", "id": str(a.id)} for a in albums]
@@ -586,7 +598,7 @@ class ArtistDocument(AURADocument):
         Args:
             artist_id: A string which is the artist's name.
         """
-        artist_resource = self.resource_object(artist_id)
+        artist_resource = self.get_resource_object(self.lib, artist_id)
         if not artist_resource:
             return self.error(
                 "404 Not Found",
@@ -598,11 +610,31 @@ class ArtistDocument(AURADocument):
         return self.single_resource_document(artist_resource)
 
 
+def safe_filename(fn):
+    """Check whether a string is a simple (non-path) filename.
+
+    For example, `foo.txt` is safe because it is a "plain" filename. But
+    `foo/bar.txt` and `../foo.txt` and `.` are all non-safe because they
+    can traverse to other directories other than the current one.
+    """
+    # Rule out any directories.
+    if os.path.basename(fn) != fn:
+        return False
+
+    # In single names, rule out Unix directory traversal names.
+    if fn in (".", ".."):
+        return False
+
+    return True
+
+
 class ImageDocument(AURADocument):
     """Class for building documents for /images/(id) endpoints."""
 
+    model_cls = Album
+
     @staticmethod
-    def get_image_path(image_id):
+    def get_image_path(lib: Library, image_id):
         """Works out the full path to the image with the given id.
 
         Returns None if there is no such image.
@@ -619,30 +651,32 @@ class ImageDocument(AURADocument):
         parent_type = id_split[0]
         parent_id = id_split[1]
         img_filename = "-".join(id_split[2:])
+        if not safe_filename(img_filename):
+            return None
 
         # Get the path to the directory parent's images are in
         if parent_type == "album":
-            album = current_app.config["lib"].get_album(int(parent_id))
+            album = lib.get_album(int(parent_id))
             if not album or not album.artpath:
                 return None
             # Cut the filename off of artpath
             # This is in preparation for supporting images in the same
             # directory that are not tracked by beets.
-            artpath = py3_path(album.artpath)
+            artpath = os.fsdecode(album.artpath)
             dir_path = "/".join(artpath.split("/")[:-1])
         else:
             # Images for other resource types are not supported
             return None
 
-        img_path = dir_path + "/" + img_filename
+        img_path = os.path.join(dir_path, img_filename)
         # Check the image actually exists
-        if isfile(img_path):
+        if os.path.isfile(img_path):
             return img_path
         else:
             return None
 
     @staticmethod
-    def resource_object(image_id):
+    def get_resource_object(lib: Library, image_id):
         """Construct a JSON:API resource object for the given image.
 
         Args:
@@ -651,14 +685,14 @@ class ImageDocument(AURADocument):
         """
         # Could be called as a static method, so can't use
         # self.get_image_path()
-        image_path = ImageDocument.get_image_path(image_id)
+        image_path = ImageDocument.get_image_path(lib, image_id)
         if not image_path:
             return None
 
         attributes = {
             "role": "cover",
             "mimetype": guess_type(image_path)[0],
-            "size": getsize(image_path),
+            "size": os.path.getsize(image_path),
         }
         try:
             from PIL import Image
@@ -691,7 +725,7 @@ class ImageDocument(AURADocument):
             image_id: A string in the form
                 "<parent_type>-<parent_id>-<img_filename>".
         """
-        image_resource = self.resource_object(image_id)
+        image_resource = self.get_resource_object(self.lib, image_id)
         if not image_resource:
             return self.error(
                 "404 Not Found",
@@ -719,8 +753,7 @@ def server_info():
 @aura_bp.route("/tracks")
 def all_tracks():
     """Respond with a list of all tracks and related information."""
-    doc = TrackDocument()
-    return doc.all_resources()
+    return TrackDocument.from_app().all_resources()
 
 
 @aura_bp.route("/tracks/<int:track_id>")
@@ -730,8 +763,7 @@ def single_track(track_id):
     Args:
         track_id: The id of the track provided in the URL (integer).
     """
-    doc = TrackDocument()
-    return doc.single_resource(track_id)
+    return TrackDocument.from_app().single_resource(track_id)
 
 
 @aura_bp.route("/tracks/<int:track_id>/audio")
@@ -751,8 +783,8 @@ def audio_file(track_id):
             ),
         )
 
-    path = py3_path(track.path)
-    if not isfile(path):
+    path = os.fsdecode(track.path)
+    if not os.path.isfile(path):
         return AURADocument.error(
             "404 Not Found",
             "No audio file for the requested track.",
@@ -803,8 +835,7 @@ def audio_file(track_id):
 @aura_bp.route("/albums")
 def all_albums():
     """Respond with a list of all albums and related information."""
-    doc = AlbumDocument()
-    return doc.all_resources()
+    return AlbumDocument.from_app().all_resources()
 
 
 @aura_bp.route("/albums/<int:album_id>")
@@ -814,8 +845,7 @@ def single_album(album_id):
     Args:
         album_id: The id of the album provided in the URL (integer).
     """
-    doc = AlbumDocument()
-    return doc.single_resource(album_id)
+    return AlbumDocument.from_app().single_resource(album_id)
 
 
 # Artist endpoints
@@ -825,8 +855,7 @@ def single_album(album_id):
 @aura_bp.route("/artists")
 def all_artists():
     """Respond with a list of all artists and related information."""
-    doc = ArtistDocument()
-    return doc.all_resources()
+    return ArtistDocument.from_app().all_resources()
 
 
 # Using the path converter allows slashes in artist_id
@@ -838,8 +867,7 @@ def single_artist(artist_id):
         artist_id: The id of the artist provided in the URL. A string
             which is the artist's name.
     """
-    doc = ArtistDocument()
-    return doc.single_resource(artist_id)
+    return ArtistDocument.from_app().single_resource(artist_id)
 
 
 # Image endpoints
@@ -855,8 +883,7 @@ def single_image(image_id):
         image_id: The id of the image provided in the URL. A string in
             the form "<parent_type>-<parent_id>-<img_filename>".
     """
-    doc = ImageDocument()
-    return doc.single_resource(image_id)
+    return ImageDocument.from_app().single_resource(image_id)
 
 
 @aura_bp.route("/images/<string:image_id>/file")
@@ -867,7 +894,7 @@ def image_file(image_id):
         image_id: The id of the image provided in the URL. A string in
             the form "<parent_type>-<parent_id>-<img_filename>".
     """
-    img_path = ImageDocument.get_image_path(image_id)
+    img_path = ImageDocument.get_image_path(current_app.config["lib"], image_id)
     if not img_path:
         return AURADocument.error(
             "404 Not Found",
@@ -886,7 +913,7 @@ def create_app():
     """An application factory for use by a WSGI server."""
     config["aura"].add(
         {
-            "host": u"127.0.0.1",
+            "host": "127.0.0.1",
             "port": 8337,
             "cors": [],
             "cors_supports_credentials": False,
@@ -932,7 +959,7 @@ class AURAPlugin(BeetsPlugin):
 
     def __init__(self):
         """Add configuration options for the AURA plugin."""
-        super(AURAPlugin, self).__init__()
+        super().__init__()
 
     def commands(self):
         """Add subcommand used to run the AURA server."""
@@ -954,13 +981,13 @@ class AURAPlugin(BeetsPlugin):
                 threaded=True,
             )
 
-        run_aura_cmd = Subcommand("aura", help=u"run an AURA server")
+        run_aura_cmd = Subcommand("aura", help="run an AURA server")
         run_aura_cmd.parser.add_option(
-            u"-d",
-            u"--debug",
+            "-d",
+            "--debug",
             action="store_true",
             default=False,
-            help=u"use Flask debug mode",
+            help="use Flask debug mode",
         )
         run_aura_cmd.func = run_aura
         return [run_aura_cmd]
